'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.Queue = undefined;

var _typeof = typeof Symbol === "function" && typeof Symbol.iterator === "symbol" ? function (obj) { return typeof obj; } : function (obj) { return obj && typeof Symbol === "function" && obj.constructor === Symbol && obj !== Symbol.prototype ? "symbol" : typeof obj; };

var _amqplib = require('amqplib');

var amqp = _interopRequireWildcard(_amqplib);

var _jobTracker = require('./job-tracker');

var jobTracker = _interopRequireWildcard(_jobTracker);

var _util = require('./util');

function _interopRequireWildcard(obj) { if (obj && obj.__esModule) { return obj; } else { var newObj = {}; if (obj != null) { for (var key in obj) { if (Object.prototype.hasOwnProperty.call(obj, key)) newObj[key] = obj[key]; } } newObj.default = obj; return newObj; } }

function _asyncToGenerator(fn) { return function () { var gen = fn.apply(this, arguments); return new Promise(function (resolve, reject) { function step(key, arg) { try { var info = gen[key](arg); var value = info.value; } catch (error) { reject(error); return; } if (info.done) { resolve(value); } else { return Promise.resolve(value).then(function (value) { step("next", value); }, function (err) { step("throw", err); }); } } return step("next"); }); }; }

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

var uuid = require('uuid/v4');
var debug = require('debug')('minion-kit:queue');

var CONNECTION_COUNT = 1;
var CHANNEL_COUNT_PER_CONNECTION = 3;
var PENDING_JOB_TTL = 60;

var Queue = exports.Queue = function Queue() {
  var _this = this;

  var setup = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {};

  _classCallCheck(this, Queue);

  this.currentChannelIdx = 0;
  this.connections = [];
  this.consumerChannels = [];
  this.channels = [];
  this.connect = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee3() {
    var conList, chList;
    return regeneratorRuntime.wrap(function _callee3$(_context3) {
      while (1) {
        switch (_context3.prev = _context3.next) {
          case 0:
            conList = Array(_this.connectionCount).fill(0);
            chList = Array(_this.channelCountPerConnection).fill(0);
            _context3.next = 4;
            return Promise.all(conList.map(function () {
              var _ref2 = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee2(con, conIdx) {
                var newConnection;
                return regeneratorRuntime.wrap(function _callee2$(_context2) {
                  while (1) {
                    switch (_context2.prev = _context2.next) {
                      case 0:
                        _context2.next = 2;
                        return amqp.connect(_this.connectionString);

                      case 2:
                        newConnection = _context2.sent;

                        _this.connections.push(newConnection);
                        debug('Established connection #' + conIdx + ' with message queue');

                        _context2.next = 7;
                        return Promise.all(chList.map(function () {
                          var _ref3 = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee(ch, chIdx) {
                            var newChannel;
                            return regeneratorRuntime.wrap(function _callee$(_context) {
                              while (1) {
                                switch (_context.prev = _context.next) {
                                  case 0:
                                    _context.next = 2;
                                    return newConnection.createConfirmChannel();

                                  case 2:
                                    newChannel = _context.sent;

                                    _this.channels.push(newChannel);
                                    debug('Opened channel #' + chIdx + ' on connection #' + conIdx + ' with message queue (total: #' + _this.channels.length + ')');

                                  case 5:
                                  case 'end':
                                    return _context.stop();
                                }
                              }
                            }, _callee, _this);
                          }));

                          return function (_x4, _x5) {
                            return _ref3.apply(this, arguments);
                          };
                        }()));

                      case 7:
                      case 'end':
                        return _context2.stop();
                    }
                  }
                }, _callee2, _this);
              }));

              return function (_x2, _x3) {
                return _ref2.apply(this, arguments);
              };
            }()));

          case 4:
          case 'end':
            return _context3.stop();
        }
      }
    }, _callee3, _this);
  }));
  this.disconnect = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee7() {
    return regeneratorRuntime.wrap(function _callee7$(_context7) {
      while (1) {
        switch (_context7.prev = _context7.next) {
          case 0:
            _context7.next = 2;
            return Promise.all(_this.channels.map(function () {
              var _ref5 = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee4(ch, chIdx) {
                return regeneratorRuntime.wrap(function _callee4$(_context4) {
                  while (1) {
                    switch (_context4.prev = _context4.next) {
                      case 0:
                        debug('Closing channel #' + chIdx);
                        _context4.next = 3;
                        return ch.waitForConfirms();

                      case 3:
                      case 'end':
                        return _context4.stop();
                    }
                  }
                }, _callee4, _this);
              }));

              return function (_x6, _x7) {
                return _ref5.apply(this, arguments);
              };
            }()));

          case 2:
            _context7.next = 4;
            return Promise.all(_this.consumerChannels.map(function () {
              var _ref6 = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee5(ch, chIdx) {
                return regeneratorRuntime.wrap(function _callee5$(_context5) {
                  while (1) {
                    switch (_context5.prev = _context5.next) {
                      case 0:
                        debug('Closing consumer channel #' + chIdx);
                        _context5.next = 3;
                        return ch.waitForConfirms();

                      case 3:
                      case 'end':
                        return _context5.stop();
                    }
                  }
                }, _callee5, _this);
              }));

              return function (_x8, _x9) {
                return _ref6.apply(this, arguments);
              };
            }()));

          case 4:
            _context7.next = 6;
            return Promise.all(_this.connections.map(function () {
              var _ref7 = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee6(con, conIdx) {
                return regeneratorRuntime.wrap(function _callee6$(_context6) {
                  while (1) {
                    switch (_context6.prev = _context6.next) {
                      case 0:
                        debug('Closing connection #' + conIdx);
                        _context6.next = 3;
                        return con.close();

                      case 3:
                      case 'end':
                        return _context6.stop();
                    }
                  }
                }, _callee6, _this);
              }));

              return function (_x10, _x11) {
                return _ref7.apply(this, arguments);
              };
            }()));

          case 6:
          case 'end':
            return _context7.stop();
        }
      }
    }, _callee7, _this);
  }));

  this.getChannel = function () {

    _this.currentChannelIdx++;

    if (_this.currentChannelIdx >= _this.channels.length) {
      _this.currentChannelIdx = 0;
    }

    return _this.channels[_this.currentChannelIdx];
  };

  this.sendToQueue = function () {
    var _ref8 = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee8(queueName, payload, options) {
      var channel;
      return regeneratorRuntime.wrap(function _callee8$(_context8) {
        while (1) {
          switch (_context8.prev = _context8.next) {
            case 0:
              channel = _this.getChannel();
              _context8.next = 3;
              return channel.assertQueue(queueName, { durable: true, noAck: false });

            case 3:

              debug('Sending new job to queue ' + queueName);

              return _context8.abrupt('return', channel.sendToQueue(queueName, Buffer.from(JSON.stringify(payload)), {
                persistent: true,
                contentType: 'application/json',
                headers: options
              }));

            case 5:
            case 'end':
              return _context8.stop();
          }
        }
      }, _callee8, _this);
    }));

    return function (_x12, _x13, _x14) {
      return _ref8.apply(this, arguments);
    };
  }();

  this.requeueJob = function () {
    var _ref9 = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee9(consumerChannel, queueName, payload, msg) {
      var finalQueueName;
      return regeneratorRuntime.wrap(function _callee9$(_context9) {
        while (1) {
          switch (_context9.prev = _context9.next) {
            case 0:
              if (!(msg.properties.headers.attemptsCount + 1 < msg.properties.headers.attemptsLimit)) {
                _context9.next = 7;
                break;
              }

              _context9.next = 3;
              return _this.sendToQueue(queueName, payload, {
                attemptsLimit: msg.properties.headers.attemptsLimit,
                attemptsCount: msg.properties.headers.attemptsCount + 1,
                finalQueueOnAttemptsExceeded: msg.properties.headers.finalQueueOnAttemptsExceeded
              });

            case 3:

              consumerChannel.nack(msg, false, false);
              debug('[' + payload.correlationId + '] Requeued job');
              _context9.next = 16;
              break;

            case 7:
              if (!msg.properties.headers.finalQueueOnAttemptsExceeded) {
                _context9.next = 15;
                break;
              }

              finalQueueName = msg.properties.headers.finalQueueOnAttemptsExceeded;
              _context9.next = 11;
              return _this.sendToQueue(finalQueueName, payload, {
                attemptsCount: msg.properties.headers.attemptsCount
              });

            case 11:

              consumerChannel.nack(msg, false, false);
              debug('[' + payload.correlationId + '] Sent job to final queue');
              _context9.next = 16;
              break;

            case 15:
              consumerChannel.nack(msg, false, false);

            case 16:
            case 'end':
              return _context9.stop();
          }
        }
      }, _callee9, _this);
    }));

    return function (_x15, _x16, _x17, _x18) {
      return _ref9.apply(this, arguments);
    };
  }();

  this.consume = function () {
    var _ref10 = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee10(consumerQueueName, jobProcessor) {
      var newConnection, consumerChannel;
      return regeneratorRuntime.wrap(function _callee10$(_context10) {
        while (1) {
          switch (_context10.prev = _context10.next) {
            case 0:
              _context10.next = 2;
              return amqp.connect(_this.connectionString);

            case 2:
              newConnection = _context10.sent;

              debug('Established consumer connection with message queue');

              _context10.next = 6;
              return newConnection.createConfirmChannel();

            case 6:
              consumerChannel = _context10.sent;

              consumerChannel.prefetch(1);
              _this.consumerChannels.push(consumerChannel);
              debug('Established consumer channel with queue ' + consumerQueueName);

              _context10.next = 12;
              return consumerChannel.assertQueue(consumerQueueName, { durable: true, noAck: false });

            case 12:

              consumerChannel.consume(consumerQueueName, function (msg) {

                var payload = void 0;

                try {
                  payload = JSON.parse(msg.content);

                  if ((typeof payload === 'undefined' ? 'undefined' : _typeof(payload)) !== 'object') {
                    throw new Error('Message should be an object');
                  }
                } catch (err) {
                  debug({ err: err }, 'Invalid message format');
                  return consumerChannel.nack(msg, false, false);
                }

                if (!payload.correlationId) {
                  payload.correlationId = uuid();
                }

                debug('[' + payload.correlationId + '] Fetched job');

                jobTracker.trackPendingJobs([payload]);

                // set defaults
                msg.properties.headers = msg.properties.headers || {};
                msg.properties.headers.attemptsLimit = msg.properties.headers.attemptsLimit || 0;
                msg.properties.headers.attemptsCount = msg.properties.headers.attemptsCount || 0;

                var jobProcessorPromise = new Promise(function (resolve, reject) {
                  try {
                    resolve(jobProcessor(payload, consumerQueueName));
                  } catch (err) {
                    reject(err);
                  }
                });

                return (0, _util.promiseWithTimeout)(_this.pendingJobTTL * 1000, jobProcessorPromise).then(function () {
                  jobTracker.increaseStatusCompleted();
                  consumerChannel.ack(msg);
                  debug('[' + payload.correlationId + '] Completed job');
                }).catch(function (err) {
                  if (err instanceof _util.TimeoutError) {
                    jobTracker.increaseStatusTimedOut();
                  } else {
                    jobTracker.increaseStatusFailed();
                  }

                  debug({ err: err, correlationId: payload.correlationId });
                  _this.requeueJob(consumerChannel, consumerQueueName, payload, msg);
                }).then(function () {
                  jobTracker.unTrackPendingJob(payload);
                });
              });

            case 13:
            case 'end':
              return _context10.stop();
          }
        }
      }, _callee10, _this);
    }));

    return function (_x19, _x20) {
      return _ref10.apply(this, arguments);
    };
  }();

  var connectionString = setup.connectionString,
      _setup$connectionCoun = setup.connectionCount,
      connectionCount = _setup$connectionCoun === undefined ? CONNECTION_COUNT : _setup$connectionCoun,
      _setup$channelCountPe = setup.channelCountPerConnection,
      channelCountPerConnection = _setup$channelCountPe === undefined ? CHANNEL_COUNT_PER_CONNECTION : _setup$channelCountPe,
      _setup$pendingJobTTL = setup.pendingJobTTL,
      pendingJobTTL = _setup$pendingJobTTL === undefined ? PENDING_JOB_TTL : _setup$pendingJobTTL;


  this.connectionString = connectionString;
  this.connectionCount = connectionCount;
  this.channelCountPerConnection = channelCountPerConnection;
  this.pendingJobTTL = pendingJobTTL;
};